import numpy as np
import matplotlib.pyplot as plt
from ase import units
import xml.etree.ElementTree as ET
import sys
from scipy.special import erf
from scipy.interpolate import interp1d

fontsize     = 15
linewidth    = 2.5
plt.rcParams.update({'font.size': fontsize, 'lines.linewidth' : linewidth})

# generate a equidistant mesh from xmin to rmax with stepsie dx
def generate_pp_r(dx, mesh, xmin, rmax, zmesh):
    return np.arange(float(xmin), float(rmax), float(dx))

# the distance between two steps is dx
def generate_pp_rab(dx, mesh, xmin, rmax, zmesh):
    return float(dx)*np.ones(int(mesh))
    
    
# potential with a smooth cutoff (caused by an exponential decay). Far behind the cutoff, the potential behaves as the potential of the He-atom for large distances, i.e. as -4/r
def pot_wo_cut(r, pot_cut, k):
    f = 0.94 # this value was optimized such that the output potential is smooth while at the same time the decay into -4/r happens fast. It is defined as r_1 = f*pot_cut, where r_1 is the starting point of the exponential decay
    #The smoothing-procedure that is applied here only works for larger enough cutoffs (otherwise the exponential decay turns into an expontial increase
    #to be on the very safe side, take pot_cut>= 2.5 Bohr radii (to avoid that "a" takes negative values)
    assert(pot_cut > (4.0/(k*f*(f+1)))**(1.0/3.0))
    r_1 = f*pot_cut
    const = 0.5*k*(pot_cut**2-r_1**2)+2.0/pot_cut-2.0/r_1
    a = k*r_1/const
    A = 2*const/np.exp(-a*r_1)
    if(r<r_1):
        return 2*0.5*k*(r**2-pot_cut**2)-2*2.0/pot_cut
    else:
        return -A*np.exp(-a*r)-2*2/r

if __name__ == '__main__':    
    
    # cut of in quantum espresso, at the moment only needed for plotting
    rcut = 12.0

    # read cutoff value of the potential from command line
    pot_cut_angstrom = float(sys.argv[1]) # in angstroms
    pot_cut = pot_cut_angstrom*1.8897259886 # convert to Bohr radii
    
    # read cell size value from command line
    cell_size_angstrom = float(sys.argv[2]) # in angstroms
    cell_size = cell_size_angstrom*1.8897259886 # convert to Bohr radii
    
    # read from command line if the PBE should also be provided in upf file -> needed for kcp.x
    if(len(sys.argv)==4):
        provide_density = True
        path_to_PBE_density = sys.argv[3] 
        # read density from file
        density_result = np.loadtxt(path_to_PBE_density)
        # generate an interpolation function to evaluate the density on an arbitrary grid
        density_func = interp1d(density_result[:,0]*cell_size, density_result[:,1])
        def density_generator(r):
            if(r<cell_size*0.5):
                return density_func(r)
            else:
                return 0.0
    else:
        provide_density = False
        

    # Use this pseudopotential as a template
    with open('../../pseudopotentials/H.coulomb-ae.UPF', 'r') as fd:
        tree = ET.parse(fd)
    upf = tree.getroot()

    pp_mesh = upf.find('PP_MESH')
    
    #for the mesh we use a linspace instead of the usual distribution: exp(xmin+i*dx)/Zmesh
    #therefore certain variables like xmin have a different meaning than usual, but in most cases, these variables are ignored by QE
    
    mesh = 3500 # = maximum number of gridpoints that is allowed by QE
    rmax = cell_size/2.0
    dx   = rmax/mesh
    xmin = 0.0 # starting point of the linspace grid
    zmesh = 2.0 #should correspond to the valence charge
    
    
    # Set the mesh attributes
    pp_mesh.attrib['mesh'] = str(mesh)
    pp_mesh.attrib['rmax'] = str(float(rmax))
    pp_mesh.attrib['dx'] = str(format(dx, "1.12e"))
    pp_mesh.attrib['xmin'] = str(format(xmin, "1.12e"))
    pp_mesh.attrib['zmesh'] = str(float(zmesh))
    
    # generate the radial grid
    pp_r_array   = generate_pp_r(**pp_mesh.attrib)
    pp_rab_array = generate_pp_rab(**pp_mesh.attrib)

    # Clear PP_INFO
    pp_info = upf.find('PP_INFO')
    pp_info.text = '\n'
    
    # Clear PP_INFO
    pp_pswfc = upf.find('PP_PSWFC')
    pp_pswfc.text = '\n'

    # Remove PP_RHOATOM
    pp_rhoatom = upf.find('PP_RHOATOM')
    if(not provide_density):
        upf.remove(pp_rhoatom)

    # Overwrite PP_LOCAL
    pp_local = upf.find('PP_LOCAL')

    ncols = 4
    mesh = int(pp_mesh.attrib['mesh'])
    pp_local.attrib = {'columns': str(ncols)}

    # Overwrite PP_R
    pp_r = pp_mesh.find('PP_R')

    # Overwrite PP_RAB
    pp_rab = pp_mesh.find('PP_RAB')


    # Modifying header
    pp_header = upf.find('PP_HEADER')
    pp_header.attrib.update(mesh_size = str(mesh), author='Edward Linscott and Yannick Schubert', comment='hookium', element='He', pseudo_type='other', z_valence='2', total_psenergy='0.0', generated='generated with a simple python script', is_coulomb="false")

    # Generating the new Hooke potential, with a cutoff at a certain radius
    k = 0.1**2  # k=m*omega**2, omega=0.5, m=1

    potential = np.zeros_like(pp_r_array)

    for i, r in enumerate(pp_r_array):
        potential[i] = pot_wo_cut(r,pot_cut, k) # -pot_wo_cut(rcut,pot_cut,k) shift to get a smooth bare output potential for pp.x


    potential_str = ''.join([f'\n{p:14.10e}' if i % ncols == 0 else f' {p:14.10e}' for i, p in enumerate(potential)]) + '\n'
    pp_r_str = ''.join([f'\n{p:14.10e}' if i % ncols == 0 else f' {p:14.10e}' for i, p in enumerate(pp_r_array)]) + '\n'
    pp_rab_str = ''.join([f'\n{p:14.10e}' if i % ncols == 0 else f' {p:14.10e}' for i, p in enumerate(pp_rab_array)]) + '\n'
    pp_local.text = potential_str
    pp_r.text = pp_r_str
    pp_rab.text = pp_rab_str
    
    
    if(provide_density): 
        rhoatom   = np.zeros_like(pp_r_array)
        for i, r in enumerate(pp_r_array):
            rhoatom[i]   = density_generator(r)
        rhoatom_str = ''.join([f'\n{p:14.10e}' if i % ncols == 0 else f' {p:14.10e}' for i, p in enumerate(rhoatom)]) + '\n' 
        pp_rhoatom.text = rhoatom_str 
        
        

    # Write to file
    with open('../../pseudopotentials/hookium.upf', 'wb') as fd:
        tree.write(fd, encoding='utf-8')


    # to also plot the exact density    
    const = 4/(25*np.pi*(61*np.pi+48*np.sqrt(5*np.pi)))

    def density_exact(r):
        return const/(800)*(200*(45+r**2)*np.exp(-r**2/5)+np.exp(-r**2/10)*np.sqrt(10*np.pi)*(2875+190*r**2+r**4+20*(175/r+50*r+r**3)*erf(r/(np.sqrt(10)))))
    
    
    # Plotting the new potential
    I = np.where((pp_r_array>0.0) & (pp_r_array < cell_size/2.0))
    fig, ax = plt.subplots(figsize = (10,10))
    ax.plot(pp_r_array[I], potential[I]/2.0, label = "generated potential")
    ax.plot(pp_r_array[I], -2/(pp_r_array[I])-0.0*pot_wo_cut(rcut,pot_cut,k), "--", label = "-2/r-potential")
    #ax.axvline([rcut], linestyle = "--",label = "rcut", color = "black")
    ax.axvline([pot_cut], linestyle = "--",label = "cutoff", color = "grey")
    if(provide_density):
        ax.plot(pp_r_array[I], -2*potential[0]*density_exact(pp_r_array[I]) + potential[0]/2.0, label = "Exact density")
        ax.plot(pp_r_array[I], -2*potential[0]*rhoatom[I] + potential[0]/2.0, label = "PBE interpolated density")
    ax.set_ylim((potential[0]/2.0*1.1, 0.5))
    ax.set_xlabel("Radial distance in Bohr radii")
    ax.set_ylabel("Potential in Hartree")
    ax.set_title("Plot of the Generated Potential")
    plt.xlim((-0.4,cell_size/2.0+0.4))
    ax.legend(loc = "upper left")
    plt.minorticks_on()
    plt.grid(which = "both")
    plt.savefig(str("postprocessing/plots/input_potential_cut_off_at_" + str(pot_cut_angstrom)+ ".png"))
    plt.show()
